from __future__ import annotations

import json
import logging
import os
import platform
import re
import shutil
import subprocess
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

import boto3

from aws_annoying.utils.platform import command_as_root, is_root, is_windows, os_release

from .errors import PluginNotInstalledError, UnsupportedPlatformError

if TYPE_CHECKING:
    from aws_annoying.utils.downloader import AbstractDownloader

logger = logging.getLogger(__name__)


class SessionManager:
    """AWS Session Manager plugin manager."""

    def __init__(self, *, session: boto3.session.Session | None = None) -> None:
        """Initialize SessionManager.

        Args:
            session: Boto3 session to use for AWS operations.
        """
        self.session = session or boto3.session.Session()

    # ------------------------------------------------------------------------
    # Installation
    # ------------------------------------------------------------------------
    def verify_installation(self) -> tuple[bool, Path | None, str | None]:
        """Verify installation of AWS Session Manager plugin.

        Returns:
            3-tuple of boolean flag indicating whether plugin installed, binary path and version string.
        """
        # Find plugin binary
        if not (binary_path := self._get_binary_path()):
            return False, None, None

        # Check version
        result_bytes = subprocess.run(  # noqa: S603
            [str(binary_path), "--version"],
            check=True,
            capture_output=True,
        )
        result = result_bytes.stdout.decode().strip()
        if not bool(re.match(r"[\d\.]+", result)):
            return False, binary_path, result

        return True, binary_path, result

    def _get_binary_path(self) -> Path | None:
        """Get the path to the session-manager-plugin binary."""
        binary_path_str = shutil.which("session-manager-plugin")
        if not binary_path_str:
            if is_windows():
                # Windows: use the default installation path
                binary_path = (
                    Path(os.environ["ProgramFiles"])  # noqa: SIM112
                    / "Amazon"
                    / "SessionManagerPlugin"
                    / "bin"
                    / "session-manager-plugin.exe"
                )
                if binary_path.is_file():
                    return binary_path.absolute()

            return None

        return Path(binary_path_str).absolute()

    def install(
        self,
        *,
        os: str | None = None,
        linux_distribution: _LinuxDistribution | None = None,
        arch: str | None = None,
        root: bool | None = None,
        downloader: AbstractDownloader,
    ) -> None:
        """Install AWS Session Manager plugin.

        Args:
            os: The operating system to install the plugin on. If `None`, will use the current operating system.
            linux_distribution: The Linux distribution to install the plugin on.
                If `None` and current `os` is `"Linux"`, will try to detect the distribution from current system.
            arch: The architecture to install the plugin on. If `None`, will use the current architecture.
            root: Whether to run the installation as root. If `None`, will check if the current user is root.
            downloader: File downloader to use for downloading the plugin.
        """
        os = os or platform.system()
        arch = arch or platform.machine()

        if os == "Windows":
            self._install_windows(downloader=downloader)
        elif os == "Darwin":
            self._install_macos(arch=arch, root=root or is_root(), downloader=downloader)
        elif os == "Linux":
            linux_distribution = linux_distribution or _detect_linux_distribution()
            self._install_linux(
                linux_distribution=linux_distribution,
                arch=arch,
                root=root or is_root(),
                downloader=downloader,
            )
        else:
            msg = f"Unsupported operating system: {os}"
            raise UnsupportedPlatformError(msg)

    def before_install(self, command: list[str]) -> None:
        """Hook to run before invoking plugin installation command."""

    # https://docs.aws.amazon.com/systems-manager/latest/userguide/install-plugin-windows.html
    def _install_windows(self, *, downloader: AbstractDownloader) -> None:
        """Install session-manager-plugin on Windows via EXE installer."""
        download_url = (
            "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/windows/SessionManagerPluginSetup.exe"
        )
        with tempfile.TemporaryDirectory() as temp_dir:
            p = Path(temp_dir)
            exe_installer = downloader.download(download_url, to=p / "SessionManagerPluginSetup.exe")
            command = [str(exe_installer), "/quiet"]
            self.before_install(command)
            subprocess.call(command, cwd=p)  # noqa: S603

    # https://docs.aws.amazon.com/systems-manager/latest/userguide/install-plugin-macos-overview.html
    def _install_macos(self, *, arch: str, root: bool, downloader: AbstractDownloader) -> None:
        """Install session-manager-plugin on macOS via signed installer."""
        # ! Intel chip will not be supported
        if arch == "x86_64":
            download_url = (
                "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/mac/session-manager-plugin.pkg"
            )
        elif arch == "arm64":
            download_url = (
                "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/mac_arm64/session-manager-plugin.pkg"
            )
        else:
            msg = f"Architecture {arch} not supported for macOS"
            raise UnsupportedPlatformError(msg)

        with tempfile.TemporaryDirectory() as temp_dir:
            p = Path(temp_dir)
            pkg_installer = downloader.download(download_url, to=p / "session-manager-plugin.pkg")

            # Run installer
            command = command_as_root(
                ["installer", "-pkg", str(pkg_installer), "-target", "/"],
                root=root,
            )
            self.before_install(command)
            subprocess.call(command, cwd=p)  # noqa: S603

            # Symlink
            command = [
                "ln",
                "-s",
                "/usr/local/sessionmanagerplugin/bin/session-manager-plugin",
                "/usr/local/bin/session-manager-plugin",
            ]
            if not root:
                command = ["sudo", *command]

            logger.info("Running %s to create symlink", " ".join(command))
            Path("/usr/local/bin").mkdir(exist_ok=True)
            subprocess.call(command, cwd=p)  # noqa: S603

    # https://docs.aws.amazon.com/systems-manager/latest/userguide/install-plugin-linux-overview.html
    def _install_linux(
        self,
        *,
        linux_distribution: _LinuxDistribution,
        arch: str,
        root: bool,
        downloader: AbstractDownloader,
    ) -> None:
        name = linux_distribution.name
        version = linux_distribution.version

        # Debian / Ubuntu
        if name in ("debian", "ubuntu"):
            logger.info("Detected Linux distribution: Debian / Ubuntu")

            # Download installer
            url_map = {
                "x86_64": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/ubuntu_64bit/session-manager-plugin.deb",
                "x86": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/ubuntu_32bit/session-manager-plugin.deb",
                "arm64": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/ubuntu_arm64/session-manager-plugin.deb",
            }
            download_url = url_map.get(arch)
            if not download_url:
                msg = f"Architecture {arch} not supported for distribution {name}"
                raise UnsupportedPlatformError(msg)

            with tempfile.TemporaryDirectory() as temp_dir:
                p = Path(temp_dir)
                deb_installer = downloader.download(download_url, to=p / "session-manager-plugin.deb")

                # Invoke installation command
                command = command_as_root(["dpkg", "--install", str(deb_installer)], root=root)
                self.before_install(command)
                subprocess.call(command, cwd=p)  # noqa: S603

        # Amazon Linux / RHEL
        elif name in ("amzn", "rhel"):
            logger.info("Detected Linux distribution: Amazon Linux / RHEL")

            # Determine package URL and package manager
            url_map = {
                "x86_64": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/linux_64bit/session-manager-plugin.rpm",
                "x86": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/linux_32bit/session-manager-plugin.rpm",
                "arm64": "https://s3.amazonaws.com/session-manager-downloads/plugin/latest/linux_arm64/session-manager-plugin.rpm",
            }
            package_url = url_map.get(arch)
            if not package_url:
                msg = f"Architecture {arch} not supported for distribution {name}"
                raise UnsupportedPlatformError(msg)

            use_yum = (
                # Amazon Linux 2
                name == "amzn" and version.startswith("2")
            ) or (
                # RHEL 7 Maipo (8 Ootpa / 9 Plow)
                name == "rhel" and "Maipo" in version
            )  # ... else use dnf
            package_manager = "yum" if use_yum else "dnf"

            # Invoke installation command
            command = command_as_root([package_manager, "install", "-y", package_url], root=root)
            self.before_install(command)
            subprocess.call(command)  # noqa: S603

        else:
            msg = f"Unsupported distribution: {name}"
            raise UnsupportedPlatformError(msg)

    # ------------------------------------------------------------------------
    # Command
    # ------------------------------------------------------------------------
    def build_command(
        self,
        target: str,
        document_name: str,
        parameters: dict[str, Any],
        reason: str | None = None,
    ) -> list[str]:
        """Build command for starting a session.

        Args:
            target: The target instance ID.
            document_name: The SSM document name to use for the session.
            parameters: The parameters to pass to the SSM document.
            reason: The reason for starting the session.

        Returns:
            The command to start the session.
        """
        is_installed, binary_path, version = self.verify_installation()
        if not is_installed:
            msg = "Session Manager plugin is not installed."
            raise PluginNotInstalledError(msg)

        ssm = self.session.client("ssm")
        response = ssm.start_session(
            Target=target,
            DocumentName=document_name,
            Parameters=parameters,
            # ? Reason is optional but it doesn't allow empty string or `None`
            **({"Reason": reason} if reason else {}),
        )

        region = self.session.region_name
        return [
            str(binary_path),
            json.dumps(response),
            region,
            "StartSession",
            self.session.profile_name,
            json.dumps({"Target": target}),
            f"https://ssm.{region}.amazonaws.com",
        ]


# ? Could be moved to utils, but didn't because it's too specific to this module
class _LinuxDistribution(NamedTuple):
    name: str
    version: str


def _detect_linux_distribution() -> _LinuxDistribution:
    """Autodetect current Linux distribution."""
    osr = os_release()
    name = osr.get("ID", "").lower()
    version = osr.get("VERSION", "")
    return _LinuxDistribution(name=name, version=version)



from __future__ import annotations

import os
import subprocess


def invoke_cli(*args: str, env: dict[str, str] | None = None) -> subprocess.CompletedProcess[str]:
    return subprocess.run(  # noqa: S603
        ["uv", "run", "aws-annoying", *args],  # noqa: S607
        check=False,
        capture_output=True,
        text=True,
        env=(env or os.environ),  # * `AWS_ENDPOINT_URL` should be inherited appropriately to use Moto or LocalStack
    )



import typer

from aws_annoying.cli.app import app

ecs_app = typer.Typer(
    no_args_is_help=True,
    help="ECS (Elastic Container Service) utility commands.",
)
app.add_typer(ecs_app, name="ecs")



from __future__ import annotations

import os
import platform
from pathlib import Path


def command_as_root(command: list[str], *, root: bool | None = None) -> list[str]:
    """Modify a command to run as root (`sudo`) if not already running as root."""
    root = root or is_root()
    if not root:
        command = ["sudo", *command]

    return command


def is_root() -> bool:
    """Check if the current user is root."""
    return os.geteuid() == 0


def os_release() -> dict[str, str]:
    """Parse `/etc/os-release` file into a dictionary."""
    content = Path("/etc/os-release").read_text()
    return {
        key.strip('"'): value.strip('"')
        for key, value in (line.split("=", 1) for line in content.splitlines() if "=" in line)
    }


def is_macos() -> bool:
    """Check if the current OS is macOS."""
    return platform.system() == "Darwin"


def is_windows() -> bool:
    """Check if the current OS is Windows."""
    return platform.system() == "Windows"



# flake8: noqa: B008
from __future__ import annotations

import logging
import os
import subprocess
from io import StringIO
from typing import NoReturn, Optional

import typer
from rich.console import Console
from rich.table import Table

from aws_annoying.variable_loader import VariableLoader

from .app import app

logger = logging.getLogger(__name__)


@app.command(
    context_settings={
        # Allow extra arguments for user provided command
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
)
def load_variables(
    ctx: typer.Context,
    *,
    arns: list[str] = typer.Option(
        [],
        metavar="ARN",
        help=(
            "ARNs of the secret or parameter to load."
            " The variables are loaded in the order of the ARNs,"
            " overwriting the variables with the same name in the order of the ARNs."
        ),
    ),
    env_prefix: Optional[str] = typer.Option(
        None,
        help="Prefix of the environment variables to load the ARNs from.",
        show_default=False,
    ),
    overwrite_env: bool = typer.Option(
        False,  # noqa: FBT003
        help="Overwrite the existing environment variables with the same name.",
    ),
    replace: bool = typer.Option(
        True,  # noqa: FBT003
        help=(
            "Replace the current process (`os.execvpe`) with the command."
            " If disabled, run the command as a `subprocess`."
        ),
    ),
) -> NoReturn:
    """Wrapper command to run command with variables from AWS resources injected as environment variables.

    This script is intended to be used in the ECS environment, where currently AWS does not support
    injecting whole JSON dictionary of secrets or parameters as environment variables directly.

    It first loads the variables from the AWS sources then runs the command with the variables injected as environment variables.

    In addition to `--arns` option, you can provide ARNs as the environment variables by providing `--env-prefix`.
    For example, if you have the following environment variables:

    ```shell
    export LOAD_AWS_CONFIG__001_app_config=arn:aws:secretsmanager:...
    export LOAD_AWS_CONFIG__002_db_config=arn:aws:ssm:...
    ```

    You can run the following command:

    ```shell
    aws-annoying load-variables --env-prefix LOAD_AWS_CONFIG__ -- ...
    ```

    The variables are loaded in the order of option provided, overwriting the variables with the same name in the order of the ARNs.
    Existing environment variables are preserved by default, unless `--overwrite-env` is provided.
    """  # noqa: E501
    command = ctx.args
    if not command:
        logger.warning("No command provided. Exiting...")
        raise typer.Exit(0)

    # Mapping of the ARNs by index (index used for ordering)
    map_arns_by_index = {str(idx): arn for idx, arn in enumerate(arns)}
    if env_prefix:
        logger.info("Loading ARNs from environment variables with prefix: %r", env_prefix)
        arns_env = {
            key.removeprefix(env_prefix): value for key, value in os.environ.items() if key.startswith(env_prefix)
        }
        logger.info("Found %d sources from environment variables.", len(arns_env))
        map_arns_by_index = arns_env | map_arns_by_index

    # Briefly show the ARNs
    table = Table("Index", "ARN")
    for idx, arn in sorted(map_arns_by_index.items()):
        table.add_row(idx, arn)

    # Workaround: The logger cannot directly handle the rich table output.
    with StringIO() as file:
        Console(file=file, emoji=False).print(table)
        table_str = file.getvalue().rstrip()
        logger.info("Summary:\n%s", table_str)

    # Retrieve the variables
    loader = VariableLoader()
    logger.info("Retrieving variables from AWS resources...")
    try:
        variables, load_stats = loader.load(map_arns_by_index)
    except Exception as exc:  # noqa: BLE001
        logger.error("Failed to load the variables: %s", exc)  # noqa: TRY400
        raise typer.Exit(1) from None

    logger.info("Retrieved %d secrets and %d parameters.", load_stats["secrets"], load_stats["parameters"])

    # Prepare the environment variables
    env = os.environ.copy()
    if overwrite_env:
        env.update(variables)
    else:
        # Update variables, preserving the existing ones
        for key, value in variables.items():
            env.setdefault(key, str(value))

    # Run the command with the variables injected as environment variables, replacing current process
    logger.info("Running the command: [bold orchid]%s[/bold orchid]", " ".join(command))
    if replace:  # pragma: no cover (not coverable)
        os.execvpe(command[0], command, env=env)  # noqa: S606
        # The above line should never return

    result = subprocess.run(command, env=env, check=False)  # noqa: S603
    raise typer.Exit(result.returncode)



import json
from typing import Any

import boto3

# Type aliases for better readability
Name = str
Value = Any
ARN = str


def create_secrets(kv: dict[Name, Value]) -> dict[Name, ARN]:
    secretsmanager = boto3.client("secretsmanager")
    secrets = {}
    for name, value in kv.items():
        secret = secretsmanager.create_secret(
            Name=name,
            SecretString=json.dumps(value),
        )
        secrets[name] = secret["ARN"]

    return secrets


def create_parameters(kv: dict[Name, Value]) -> dict[Name, ARN]:
    ssm = boto3.client("ssm")
    parameters = {}
    for name, value in kv.items():
        ssm.put_parameter(
            Name=name,
            Value=json.dumps(value),
            Type="String",
        )
        parameter = ssm.get_parameter(Name=name)["Parameter"]
        parameters[name] = parameter["ARN"]

    return parameters



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import boto3

if TYPE_CHECKING:
    from .common import ECSServiceRef

logger = logging.getLogger(__name__)


def check_service_task_definition(
    service_ref: ECSServiceRef,
    *,
    session: boto3.session.Session | None = None,
    expect: str,
) -> tuple[bool, str]:
    """Check the service's current task definition matches the expected one.

    Args:
        service_ref: The ECS service reference containing the cluster and service names.
        session: The boto3 session to use for the ECS client.
        expect: The ARN of expected task definition.

    Returns:
        A tuple containing a boolean indicating whether the task definition matches the expected one
        and the current task definition ARN.
    """
    session = session or boto3.session.Session()
    ecs = session.client("ecs")

    service_detail = ecs.describe_services(cluster=service_ref.cluster, services=[service_ref.service])["services"][0]
    current_task_definition_arn = service_detail["taskDefinition"]
    if current_task_definition_arn != expect:
        return (False, current_task_definition_arn)

    return (True, current_task_definition_arn)



from __future__ import annotations

from configparser import ConfigParser
from typing import TYPE_CHECKING

import pytest
from moto import mock_aws
from moto.server import ThreadedMotoServer
from testcontainers.localstack import LocalStackContainer
from typer.testing import CliRunner

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

runner = CliRunner()


@pytest.fixture
def set_terminal_width() -> int:
    """Set console width."""
    return 200


@pytest.fixture(autouse=True)
def patch_terminal_width(monkeypatch: pytest.MonkeyPatch, set_terminal_width: int) -> None:
    """Patch the console width."""
    monkeypatch.setenv("COLUMNS", str(set_terminal_width))


# AWS
# ----------------------------------------------------------------------------
@pytest.fixture(autouse=True)
def aws_credentials(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    """Mock AWS Credentials."""
    aws_credentials = tmp_path / "credentials"
    ini = ConfigParser()
    ini["default"] = {
        "aws_access_key_id": "testing",
        "aws_secret_access_key": "testing",
        "aws_session_token": "testing",
        "aws_security_token": "testing",
        "region": "us-east-1",
    }
    with aws_credentials.open("w") as f:
        ini.write(f)

    monkeypatch.setenv("AWS_SHARED_CREDENTIALS_FILE", str(aws_credentials))
    monkeypatch.setenv("AWS_CONFIG_FILE", str(tmp_path / "config"))

    # Intentionally malform `AWS_ENDPOINT_URL` environment variable to prevent running tests from the user environment
    # which highly likely malicious to the user's AWS account.
    monkeypatch.setenv("AWS_ENDPOINT_URL", "http://aws-not-configured:wrong-port")


# Moto
# ----------------------------------------------------------------------------
@pytest.fixture
def use_moto(monkeypatch: pytest.MonkeyPatch, aws_credentials: None) -> Iterator[None]:  # noqa: ARG001
    """Mock all AWS interactions."""
    # Also, Moto does not work well with existing LocalStack; so unset `AWS_ENDPOINT_URL`
    monkeypatch.delenv("AWS_ENDPOINT_URL", raising=False)
    with mock_aws():
        yield


# Moto Server
# ----------------------------------------------------------------------------
@pytest.fixture(scope="module")
def moto_server() -> Iterator[str]:
    """Run a Moto server for AWS mocking."""
    server = ThreadedMotoServer()
    server.start()
    host, port = server.get_host_and_port()
    yield f"http://{host}:{port}"
    server.stop()


@pytest.fixture
def use_moto_server(monkeypatch: pytest.MonkeyPatch, moto_server: str) -> None:
    """Use Moto server for AWS mocking."""
    monkeypatch.setenv("AWS_ENDPOINT_URL", moto_server)


# LocalStack
# ----------------------------------------------------------------------------
# https://testcontainers.com/guides/getting-started-with-testcontainers-for-python/
@pytest.fixture
def localstack(request: pytest.FixtureRequest) -> str:
    """Run Localstack for AWS mocking."""
    container = LocalStackContainer(image="localstack/localstack:4")
    container.start()

    def teardown() -> None:
        container.stop()

    request.addfinalizer(teardown)  # noqa: PT021
    return container.get_url()  # type: ignore[no-any-return]


@pytest.fixture
def use_localstack(monkeypatch: pytest.MonkeyPatch, localstack: str) -> None:
    """Use Localstack for AWS mocking."""
    monkeypatch.setenv("AWS_ENDPOINT_URL", localstack)



from __future__ import annotations

import re


def normalize_console_output(output: str, *, replace: dict[str, str] | None = None) -> str:
    """Normalize the console output for easier comparison."""
    # Remove leading and trailing spaces
    output = output.strip()

    # Unwrap each line
    output = re.sub(r"[ ]+\n", " ", output)

    # Extra replacements; e.g. temporary paths that may vary
    if replace:
        for old, new in replace.items():
            output = output.replace(old, new)

    # Handle Windows path separator
    output = output.replace("\\", "/")

    return output  # noqa: RET504



from __future__ import annotations

import pytest
from typer.testing import CliRunner

from aws_annoying.cli.main import app
from aws_annoying.session_manager import SessionManager
from tests._helpers import ci_only, run_if_macos, run_if_windows, skip_if_macos, skip_if_windows

runner = CliRunner()

pytestmark = [
    pytest.mark.integration,
    ci_only,
]


@skip_if_macos
@skip_if_windows
def test_install_linux() -> None:
    # Arrange
    session_manager = SessionManager()
    assert session_manager.verify_installation() == (False, None, None)

    # Act
    result = runner.invoke(app, ["session-manager", "install", "--yes"])

    # Assert
    assert result.exit_code == 0, result.stdout
    is_installed, binary_path, version = session_manager.verify_installation()
    assert is_installed is True
    assert binary_path
    assert binary_path.is_file()
    assert version is not None


@run_if_macos
def test_install_macos() -> None:
    # Arrange
    session_manager = SessionManager()
    assert session_manager.verify_installation() == (False, None, None)

    # Act
    result = runner.invoke(app, ["session-manager", "install", "--yes"])

    # Assert
    assert result.exit_code == 0, result.stdout
    is_installed, binary_path, version = session_manager.verify_installation()
    assert is_installed is True
    assert binary_path
    assert binary_path.is_file()
    assert version is not None


@run_if_windows
def test_install_windows() -> None:
    # Arrange
    session_manager = SessionManager()
    assert session_manager.verify_installation() == (False, None, None)

    # Act
    result = runner.invoke(app, ["session-manager", "install", "--yes"])

    # Assert
    assert result.exit_code == 0, result.stdout
    is_installed, binary_path, version = session_manager.verify_installation()
    assert is_installed is True
    assert binary_path
    assert binary_path.is_file()
    assert version is not None



import os

import pytest

from aws_annoying.utils.platform import is_macos, is_windows

# Test runner environment
ci_only = pytest.mark.skipif(not os.environ.get("CI", None), reason="This test only run in CI.")

# macOS-specific
skip_if_macos = pytest.mark.skipif(is_macos(), reason="Test is skipped on macOS.")
run_if_macos = pytest.mark.skipif(not is_macos(), reason="Test run only on macOS.")

# Windows-specific
skip_if_windows = pytest.mark.skipif(is_windows(), reason="Test is skipped on Windows OS.")
run_if_windows = pytest.mark.skipif(not is_windows(), reason="Test run only on Windows OS.")



from __future__ import annotations

from unittest import mock

import boto3
import pytest
from botocore.stub import Stubber

from aws_annoying.ecs import ECSServiceRef, check_service_task_definition

pytestmark = [
    pytest.mark.unit,
]


class Test_check_service_task_definition:
    def test_success(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "describe_services",
            {
                "services": [
                    {"taskDefinition": "arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:1"},
                ],
            },
            expected_params={
                "cluster": "my-cluster",
                "services": ["my-service"],
            },
        )

        # Act
        with stubber:
            ok, actual = check_service_task_definition(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                expect="arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:1",
                session=mocked_session,
            )

        # Assert
        assert ok is True
        assert actual == "arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:1"

    def test_not_matching(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "describe_services",
            {
                "services": [
                    {"taskDefinition": "arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:2"},
                ],
            },
            expected_params={
                "cluster": "my-cluster",
                "services": ["my-service"],
            },
        )

        # Act
        with stubber:
            ok, actual = check_service_task_definition(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                expect="arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:1",
                session=mocked_session,
            )

        # Assert
        assert ok is False
        assert actual == "arn:aws:ecs:ap-northeast-2:000000000000:task-definition/my-task-def:2"



from __future__ import annotations

from typing import Any

import typer
from rich.prompt import Confirm

from aws_annoying.session_manager import SessionManager as _SessionManager


# Custom session manager with console interactivity
class SessionManager(_SessionManager):
    def before_install(self, command: list[str]) -> None:
        if self._confirm:
            return

        confirm = Confirm.ask(f"Will run the following command: [bold red]{' '.join(command)}[/bold red]. Proceed?")
        if not confirm:
            raise typer.Abort

    def install(self, *args: Any, confirm: bool = False, **kwargs: Any) -> None:
        self._confirm = confirm
        return super().install(*args, **kwargs)



from __future__ import annotations

import configparser
import logging
from pathlib import Path  # noqa: TC003
from typing import Optional

from pydantic import BaseModel, ConfigDict

logger = logging.getLogger(__name__)


class MfaConfig(BaseModel):
    """MFA configuration for AWS profiles."""

    model_config = ConfigDict(extra="ignore")

    mfa_profile: Optional[str] = None
    mfa_source_profile: Optional[str] = None
    mfa_serial_number: Optional[str] = None

    def save_ini_file(self, path: Path, section_key: str) -> None:
        """Save configuration to an AWS config file."""
        config_ini = configparser.ConfigParser()
        config_ini.read(path)
        config_ini.setdefault(section_key, {})
        for k, v in self.model_dump(exclude_none=True).items():
            config_ini[section_key][k] = v

        with path.open("w") as f:
            config_ini.write(f)

        logger.debug("Saved config to %s with section %s", path, section_key)

    @classmethod
    def from_ini_file(cls, path: Path, section_key: str) -> tuple[MfaConfig, bool]:
        """Load configuration from an AWS config file, with boolean indicating if the config already exists."""
        logger.debug("Loading config from %s with section %s", path, section_key)
        config_ini = configparser.ConfigParser()
        config_ini.read(path)
        if config_ini.has_section(section_key):
            section = dict(config_ini.items(section_key))
            return cls.model_validate(section), True

        return cls(), False


def update_credentials(path: Path, profile: str, *, access_key: str, secret_key: str, session_token: str) -> None:
    """Update AWS credentials file with the provided profile and credentials."""
    credentials_ini = configparser.ConfigParser()
    credentials_ini.read(path)
    credentials_ini.setdefault(profile, {})
    credentials_ini[profile]["aws_access_key_id"] = access_key
    credentials_ini[profile]["aws_secret_access_key"] = secret_key
    credentials_ini[profile]["aws_session_token"] = session_token
    with path.open("w") as f:
        credentials_ini.write(f)

    logger.debug("Updated credentials file %s with profile %s", path, profile)



from __future__ import annotations

import re

import boto3
import pytest

from aws_annoying.utils.ec2 import get_instance_id_by_name

pytestmark = [
    pytest.mark.unit,
    pytest.mark.usefixtures("use_moto"),
]


class Test_get_instance_id_by_name:
    def test_get_by_name(self) -> None:
        # Arrange
        ec2 = boto3.client("ec2")
        ec2.run_instances(
            ImageId="ami-12345678",
            InstanceType="t2.micro",
            MinCount=1,
            MaxCount=1,
            TagSpecifications=[{"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": "my-instance"}]}],
        )

        # Act
        instance_id = get_instance_id_by_name("my-instance")

        # Assert
        assert len(ec2.describe_instances()["Reservations"]) == 1
        assert instance_id is not None
        assert re.match(r"^i-[0-9a-f]+$", instance_id) is not None

    def test_get_by_name_not_found(self) -> None:
        # Arrange
        ec2 = boto3.client("ec2")

        # Act
        instance_id = get_instance_id_by_name("my-instance")

        # Assert
        assert len(ec2.describe_instances()["Reservations"]) == 0
        assert instance_id is None

    def test_get_by_instance_id(self) -> None:
        # Arrange
        ec2 = boto3.client("ec2")
        response = ec2.run_instances(
            ImageId="ami-12345678",
            InstanceType="t2.micro",
            MinCount=1,
            MaxCount=1,
            TagSpecifications=[{"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": "my-instance"}]}],
        )
        instance_id = response["Instances"][0]["InstanceId"]

        # Act
        fetched_instance_id = get_instance_id_by_name(instance_id)

        # Assert
        assert len(ec2.describe_instances()["Reservations"]) == 1
        assert fetched_instance_id == instance_id

    def test_if_multiple_instances_exists(self) -> None:
        # Arrange
        ec2 = boto3.client("ec2")
        for _ in range(3):
            ec2.run_instances(
                ImageId="ami-12345678",
                InstanceType="t2.micro",
                MinCount=1,
                MaxCount=1,
                TagSpecifications=[{"ResourceType": "instance", "Tags": [{"Key": "Name", "Value": "my-instance"}]}],
            )

        # Act
        instance_id = get_instance_id_by_name("my-instance")

        # Assert
        assert len(ec2.describe_instances()["Reservations"]) == 3
        assert instance_id is not None
        assert re.match(r"^i-[0-9a-f]+$", instance_id) is not None



from __future__ import annotations

import logging
import subprocess
from contextlib import contextmanager
from typing import TYPE_CHECKING

from aws_annoying.utils.timeout import Timeout

from .session_manager import SessionManager

if TYPE_CHECKING:
    from collections.abc import Iterator


logger = logging.getLogger(__name__)


@contextmanager
def port_forward(  # noqa: PLR0913
    *,
    through: str,
    local_port: int,
    remote_host: str,
    remote_port: int,
    reason: str | None = None,
    start_timeout: int | None = None,
) -> Iterator[subprocess.Popen[str]]:
    """Context manager for port forwarding sessions.

    Args:
        through: The instance ID to use as port-forwarding proxy.
        local_port: The local port to listen to.
        remote_host: The remote host to connect to.
        remote_port: The remote port to connect to.
        reason: The reason for starting the session.
        start_timeout: The timeout in seconds to wait for the session to start.

    Returns:
        The command to start the session.
    """
    session_manager = SessionManager()
    command = session_manager.build_command(
        target=through,
        document_name="AWS-StartPortForwardingSessionToRemoteHost",
        parameters={
            "localPortNumber": [str(local_port)],
            "host": [remote_host],
            "portNumber": [str(remote_port)],
        },
        reason=reason,
    )
    try:
        proc = subprocess.Popen(  # noqa: S603
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
        )

        # * Must be unreachable
        if proc.stdout is None:
            msg = "Standard output is not available"
            raise RuntimeError(msg)

        # Wait for the session to start
        # ? Not sure this is trustworthy health check
        with Timeout(start_timeout):
            for line in proc.stdout:
                if "Waiting for connections..." in line:
                    logger.info("Session started successfully.")
                    break

        yield proc
    finally:
        proc.terminate()



from pathlib import Path

from .boto3 import create_parameters, create_secrets
from .command_builder import repeat_options
from .invoke import invoke_cli
from .string_ import normalize_console_output

PRINTENV_PY = (Path(__file__).parent / "scripts" / "printenv.py").absolute()

__all__ = (
    "PRINTENV_PY",
    "create_parameters",
    "create_secrets",
    "invoke_cli",
    "normalize_console_output",
    "repeat_options",
)



from __future__ import annotations

import logging
import os
import signal
from pathlib import Path  # noqa: TC003

import typer

from ._app import session_manager_app

logger = logging.getLogger(__name__)


@session_manager_app.command()
def stop(
    ctx: typer.Context,
    *,
    pid_file: Path = typer.Option(  # noqa: B008
        "./session-manager-plugin.pid",
        help="The path to the PID file to store the process ID of the session manager plugin.",
    ),
    remove: bool = typer.Option(
        True,  # noqa: FBT003
        help="Remove the PID file after stopping the session.",
    ),
) -> None:
    """Stop running session for PID file."""
    dry_run = ctx.meta["dry_run"]

    # Check if PID file exists
    if not pid_file.is_file():
        logger.error("PID file not found: %s", pid_file)
        raise typer.Exit(1)

    # Read PID from file
    pid_content = pid_file.read_text()
    try:
        pid = int(pid_content)
    except ValueError:
        logger.error("PID file content is invalid; expected integer, but got: %s", type(pid_content))  # noqa: TRY400
        raise typer.Exit(1) from None

    # Send SIGTERM to the process
    try:
        logger.warning("Terminating running process with PID %d.", pid)
        if not dry_run:
            os.kill(pid, signal.SIGTERM)
    except ProcessLookupError:
        logger.warning("Tried to terminate process with PID %d but does not exist.", pid)

    # Remove the PID file
    if remove:
        logger.info("Removed the PID file %s.", pid_file)
        if not dry_run:
            pid_file.unlink()

    logger.info("Terminated the session successfully.")



from __future__ import annotations

from configparser import ConfigParser
from typing import TYPE_CHECKING
from unittest import mock

import pytest
from typer.testing import CliRunner

from aws_annoying.cli.main import app
from aws_annoying.mfa_config import MfaConfig
from tests.cli._helpers import normalize_console_output

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_snapshot.plugin import Snapshot

runner = CliRunner()

pytestmark = [
    pytest.mark.unit,
    pytest.mark.usefixtures("use_moto"),
]


@pytest.mark.parametrize("skip_persist", [True, False], ids=["skip_persist", "persist"])
def test_basic(snapshot: Snapshot, tmp_path: Path, skip_persist: bool) -> None:  # noqa: FBT001
    """The command should configure MFA settings."""
    # Arrange
    mfa_profile = "mfa"
    aws_credentials = tmp_path / "credentials"
    aws_config = tmp_path / "config"

    # Act
    result = runner.invoke(
        app,
        [
            "mfa",
            "configure",
            "--mfa-profile",
            mfa_profile,
            "--mfa-source-profile",
            "default",
            "--mfa-serial-number",
            "1234567890",
            "--mfa-token-code",
            "123456",
            "--aws-credentials",
            str(aws_credentials),
            "--aws-config",
            str(aws_config),
            *(["--no-persist"] if skip_persist else []),
        ],
    )

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(
        normalize_console_output(result.stdout, replace={str(tmp_path): "<tmp_path>"}),
        "stdout.txt",
    )

    ini = ConfigParser()
    ini.read(aws_credentials)
    assert ini[mfa_profile] == {
        "aws_access_key_id": mock.ANY,
        "aws_secret_access_key": mock.ANY,
        "aws_session_token": mock.ANY,
    }

    if skip_persist:
        assert not aws_config.exists()
    else:
        snapshot.assert_match(aws_config.read_text(), "aws_config.ini")


def test_load_existing_config(snapshot: Snapshot, tmp_path: Path) -> None:
    """The command should load existing config if arguments not given."""
    # Arrange
    mfa_profile = "mfa"
    aws_credentials = tmp_path / "credentials"
    aws_config = tmp_path / "config"
    MfaConfig(
        mfa_profile=mfa_profile,
        mfa_source_profile="default",
        mfa_serial_number="1234567890",
    ).save_ini_file(aws_config, "aws-annoying:mfa")

    # Act
    result = runner.invoke(
        app,
        [
            "mfa",
            "configure",
            "--aws-credentials",
            str(aws_credentials),
            "--aws-config",
            str(aws_config),
            "--mfa-token-code",
            "123456",
        ],
    )

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(
        normalize_console_output(result.stdout, replace={str(tmp_path): "<tmp_path>"}),
        "stdout.txt",
    )

    ini = ConfigParser()
    ini.read(aws_credentials)
    assert ini[mfa_profile] == {
        "aws_access_key_id": mock.ANY,
        "aws_secret_access_key": mock.ANY,
        "aws_session_token": mock.ANY,
    }

    snapshot.assert_match(aws_config.read_text(), "aws_config.ini")


def test_dry_run(snapshot: Snapshot, tmp_path: Path) -> None:
    """If dry-run mode enabled, configuration shouldn't updated."""
    # Arrange
    mfa_profile = "mfa"
    aws_credentials = tmp_path / "credentials"
    aws_config = tmp_path / "config"

    # Act
    result = runner.invoke(
        app,
        [
            "--dry-run",
            "mfa",
            "configure",
            "--mfa-profile",
            mfa_profile,
            "--mfa-source-profile",
            "default",
            "--mfa-serial-number",
            "1234567890",
            "--mfa-token-code",
            "123456",
            "--aws-credentials",
            str(aws_credentials),
            "--aws-config",
            str(aws_config),
        ],
    )

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(
        normalize_console_output(result.stdout, replace={str(tmp_path): "<tmp_path>"}),
        "stdout.txt",
    )

    ini = ConfigParser()
    ini.read(aws_credentials)
    assert mfa_profile not in ini
    assert not aws_config.exists()



from __future__ import annotations

from typing import TYPE_CHECKING

import boto3
import pytest
from typer.testing import CliRunner

from aws_annoying.cli.main import app
from tests.cli._helpers import normalize_console_output

if TYPE_CHECKING:
    from pytest_snapshot.plugin import Snapshot

runner = CliRunner()

pytestmark = [
    pytest.mark.unit,
    pytest.mark.usefixtures("use_moto"),
]


def test_basic(snapshot: Snapshot) -> None:
    """The command should deregister the oldest task definitions."""
    # Arrange
    ecs = boto3.client("ecs")
    family = "my-task"
    num_task_defs = 25
    for i in range(1, num_task_defs + 1):
        ecs.register_task_definition(
            family=family,
            containerDefinitions=[
                {
                    "name": "my-container",
                    "image": f"my-image:{i}",
                    "cpu": 0,
                    "memory": 0,
                },
            ],
        )

    # Act
    keep_latest = 10
    result = runner.invoke(
        app,
        [
            "ecs",
            "task-definition-lifecycle",
            "--family",
            family,
            "--keep-latest",
            str(keep_latest),
            # ? `delete_task_definitions` not implemented in moto yet
            # "--delete",
        ],
    )
    task_definitions = [
        ecs.describe_task_definition(taskDefinition=f"{family}:{i}")["taskDefinition"]
        for i in range(1, num_task_defs + 1)
    ]

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")

    # ?: Moto (v5.1.1) `ecs.list_task_definitions` does not handle `status` filter properly
    # ?:               + sorting also does not work (current behavior is ASC)
    assert [td["revision"] for td in task_definitions if td["status"] == "INACTIVE"] == list(
        range(1, num_task_defs - keep_latest + 1),  # 1..15
    )
    assert [td["revision"] for td in task_definitions if td["status"] == "ACTIVE"] == list(
        range(num_task_defs - keep_latest + 1, num_task_defs + 1),  # 16..25
    )


def test_dry_run(snapshot: Snapshot) -> None:
    """If `--dry-run` option given, the command should not perform any changes."""
    # Arrange
    ecs = boto3.client("ecs")
    family = "my-task"
    num_task_defs = 25
    for i in range(1, num_task_defs + 1):
        ecs.register_task_definition(
            family=family,
            containerDefinitions=[
                {
                    "name": "my-container",
                    "image": f"my-image:{i}",
                    "cpu": 0,
                    "memory": 0,
                },
            ],
        )

    # Act
    keep_latest = 10
    result = runner.invoke(
        app,
        [
            "--dry-run",
            "ecs",
            "task-definition-lifecycle",
            "--family",
            family,
            "--keep-latest",
            str(keep_latest),
            "--delete",
        ],
    )
    task_definitions = [
        ecs.describe_task_definition(taskDefinition=f"{family}:{i}")["taskDefinition"]
        for i in range(1, num_task_defs + 1)
    ]

    # Assert
    assert result.exit_code == 0
    snapshot.assert_match(normalize_console_output(result.stdout), "stdout.txt")

    # ?: Moto (v5.1.1) `ecs.list_task_definitions` does not handle `status` filter properly
    # ?:               + sorting also does not work (current behavior is ASC)
    assert [td["revision"] for td in task_definitions if td["status"] == "INACTIVE"] == []
    assert len([td["revision"] for td in task_definitions if td["status"] == "ACTIVE"]) == num_task_defs



from __future__ import annotations

import signal
from functools import wraps
from typing import TYPE_CHECKING, Callable, TypeVar, cast

from aws_annoying.utils.platform import is_windows

if TYPE_CHECKING:
    from types import FrameType
    from typing import Any


class OperationTimeoutError(Exception):
    """Operation timed out."""


_F = TypeVar("_F", bound=Callable)


class Timeout:
    """Timeout handler utilizing signals.

    This utility relies on Unix signals (`signal.SIGALRM`). The behavior will be dummied
    to do nothing on Windows OS.
    """

    def __init__(self, seconds: int | None) -> None:
        """Initialize timeout handler.

        Args:
            seconds: The timeout in seconds. `None` means no timeout,
                allowing the function to run normally.

        """
        self.timeout_seconds = seconds

        self._signal_handler_registered = False

    def _set_signal_handler(self) -> None:
        if is_windows() or self.timeout_seconds is None:
            return

        signal.signal(signal.SIGALRM, self._handler)
        signal.alarm(self.timeout_seconds)
        self._signal_handler_registered = True

    def _handler(self, signum: int, frame: FrameType | None) -> Any:  # noqa: ARG002
        msg = "Timeout reached"
        raise OperationTimeoutError(msg)

    def _reset_signal_handler(self) -> None:
        if is_windows() or not self._signal_handler_registered:
            return

        signal.signal(signal.SIGALRM, signal.SIG_IGN)
        signal.alarm(0)
        self._signal_handler_registered = False

    def __call__(self, func: _F) -> _F:
        """Decorator to set a timeout for a function.

        Please note, using this decorator in nested functions may not work properly as
        the signal handler for outer functions may not be resumed correctly.

        Raises:
            OperationTimeoutError: When timeout is reached.
        """

        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            self._set_signal_handler()
            try:
                return func(*args, **kwargs)
            finally:
                self._reset_signal_handler()

        return cast("_F", wrapper)

    def __enter__(self) -> None:
        self._set_signal_handler()

    def __exit__(self, *args: object) -> Any:
        self._reset_signal_handler()
        return False  # Re-raise



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import requests
from tqdm import tqdm

if TYPE_CHECKING:
    from pathlib import Path

logger = logging.getLogger(__name__)


class AbstractDownloader(ABC):
    """Abstract downloader class for downloading files."""

    @abstractmethod
    def download(self, url: str, *, to: Path) -> Path:
        """Download file from URL to path."""


class DummyDownloader(AbstractDownloader):
    """Dummy downloader that does nothing (mainly for testing purposes)."""

    def download(self, url: str, *, to: Path) -> Path:
        """Download file from URL to path."""
        logger.debug("Dummy downloader called for URL (%s) to %s.", url, to)
        return to.absolute()


class TQDMDownloader(AbstractDownloader):
    """Downloader with TQDM progress bar."""

    def download(self, url: str, *, to: Path) -> Path:
        """Download file from URL to path."""
        # https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
        logger.info("Downloading file from URL (%s) to %s.", url, to)
        with requests.get(url, stream=True, timeout=10) as response:
            response.raise_for_status()
            total_size = int(response.headers.get("content-length", 0))
            with (
                to.open("wb") as f,
                tqdm(desc=url, total=total_size, unit="iB", unit_scale=True, unit_divisor=1_024) as pbar,
            ):
                for chunk in response.iter_content(chunk_size=8_192):
                    size = f.write(chunk)
                    pbar.update(size)

        return to.absolute()



from __future__ import annotations

import importlib.metadata
import logging
import logging.config
from typing import Optional

import typer
from rich import print  # noqa: A004
from rich.console import Console
from rich.highlighter import ReprHighlighter
from rich.theme import Theme

logger = logging.getLogger(__name__)

app = typer.Typer(
    pretty_exceptions_short=True,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="rich",
    no_args_is_help=True,
)


def show_version(value: Optional[bool]) -> None:
    """Show the version of the application."""
    if not value:
        return

    print(importlib.metadata.version("aws-annoying"))
    raise typer.Exit(0)


@app.callback()
def main(  # noqa: D103
    ctx: typer.Context,
    *,
    version: Optional[bool] = typer.Option(  # noqa: ARG001
        None,
        "--version",
        is_eager=True,
        callback=show_version,
        help="Show the version and exit.",
    ),
    quiet: bool = typer.Option(
        False,  # noqa: FBT003
        help="Disable outputs.",
    ),
    verbose: bool = typer.Option(
        False,  # noqa: FBT003
        help="Enable verbose outputs.",
    ),
    dry_run: bool = typer.Option(
        False,  # noqa: FBT003
        help="Enable dry-run mode. If enabled, certain commands will avoid making changes.",
    ),
) -> None:
    log_level = logging.DEBUG if verbose else logging.INFO
    console = _get_console()
    logging_config: logging.config._DictConfigArgs = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "rich": {
                "format": "%(message)s",
                "datefmt": "[%X]",
            },
        },
        "handlers": {
            "null": {
                "class": "logging.NullHandler",
            },
            "rich": {
                "class": "aws_annoying.cli.logging_handler.RichLogHandler",
                "formatter": "rich",
                "console": console,
            },
        },
        "root": {
            "handlers": ["null"],
        },
        "loggers": {
            "aws_annoying": {
                "level": log_level,
                "handlers": ["rich"],
                "propagate": True,
            },
        },
    }
    if quiet:
        logging_config["loggers"]["aws_annoying"]["level"] = logging.CRITICAL

    logging.config.dictConfig(logging_config)

    # Global flags
    ctx.meta["dry_run"] = dry_run
    if dry_run:
        logger.warning("Dry run mode enabled. Some operation may behave differently to avoid making changes.")


def _get_console() -> Console:
    theme = Theme(
        {
            "repr.arn": "bold orange3",
            "repr.constant": "bold blue",
        },
    )
    return Console(soft_wrap=True, emoji=False, highlighter=CustomHighlighter(), theme=theme)


class CustomHighlighter(ReprHighlighter):
    """Custom highlighter to handle additional patterns."""

    highlights = [  # noqa: RUF012
        *ReprHighlighter.highlights,
        # AWS Resource Name; https://docs.aws.amazon.com/IAM/latest/UserGuide/reference-arns.html
        # NOTE: Quite simplified regex, may not cover all cases.
        r"(?P<arn>\barn:[0-9a-zA-Z/+=,\.@_\-:]+\b)",
        # Constants
        r"(?P<constant>\b[A-Z_]+\b)",
    ]



from __future__ import annotations

import logging
from datetime import datetime, timezone
from typing import Optional

import typer

from aws_annoying.ecs import (
    DeploymentFailedError,
    ECSServiceRef,
    ServiceTaskDefinitionAssertionError,
    check_service_task_definition,
    wait_for_deployment_complete,
    wait_for_deployment_start,
    wait_for_service_stability,
)
from aws_annoying.utils.timeout import OperationTimeoutError, Timeout

from ._app import ecs_app

logger = logging.getLogger(__name__)


@ecs_app.command()
def wait_for_deployment(  # noqa: PLR0913
    *,
    cluster: str = typer.Option(
        ...,
        help="The name of the ECS cluster.",
        show_default=False,
    ),
    service: str = typer.Option(
        ...,
        help="The name of the ECS service.",
        show_default=False,
    ),
    expected_task_definition: Optional[str] = typer.Option(
        None,
        help=(
            "The service's task definition expected after deployment."
            " If provided, it will be used to assert the service's task definition after deployment finished or timed out."  # noqa: E501
        ),
        show_default=False,
    ),
    polling_interval: int = typer.Option(
        5,
        help="The interval between any polling attempts, in seconds.",
        min=1,
    ),
    timeout_seconds: Optional[int] = typer.Option(
        None,
        help=(
            "The maximum time to wait for the deployment to complete, in seconds."
            " If not provided, it will wait indefinitely."
        ),
        min=1,
    ),
    wait_for_start: bool = typer.Option(
        True,  # noqa: FBT003
        help=(
            "Whether to wait for the deployment to start."
            " Because there could be no deployment right after the deploy,"
            " this option will wait for a new deployment to start if no running deployment is found."
        ),
    ),
    wait_for_stability: bool = typer.Option(
        False,  # noqa: FBT003
        help="Whether to wait for the service to be stable after the deployment.",
    ),
) -> None:
    """Wait for ECS deployment to complete."""
    start = datetime.now(tz=timezone.utc)
    try:
        with Timeout(timeout_seconds):
            _wait_for_deployment(
                ECSServiceRef(cluster=cluster, service=service),
                wait_for_start=wait_for_start,
                polling_interval=polling_interval,
                wait_for_stability=wait_for_stability,
                expected_task_definition=expected_task_definition,
            )
    except OperationTimeoutError:
        logger.error(  # noqa: TRY400
            "Timeout reached after %s seconds. The deployment may not have finished.",
            timeout_seconds,
        )
        raise typer.Exit(1) from None
    except DeploymentFailedError as err:
        elapsed = datetime.now(tz=timezone.utc) - start
        logger.error(  # noqa: TRY400
            "Deployment failed in [bold]%.2f[/bold] seconds with error: %s",
            elapsed.total_seconds(),
            err,
        )
        raise typer.Exit(1) from None
    else:
        elapsed = datetime.now(tz=timezone.utc) - start
        logger.info(
            "Deployment completed in [bold]%.2f[/bold] seconds.",
            elapsed.total_seconds(),
        )


def _wait_for_deployment(
    service_ref: ECSServiceRef,
    *,
    wait_for_start: bool,
    polling_interval: int = 5,
    wait_for_stability: bool,
    expected_task_definition: str | None = None,
) -> None:
    # Find current deployment for the service
    logger.info(
        "Looking up running deployment for service %s",
        service_ref.service,
    )
    latest_deployment_arn = wait_for_deployment_start(
        service_ref,
        wait_for_start=wait_for_start,
        polling_interval=polling_interval,
    )

    # Polling for the deployment to finish (successfully or unsuccessfully)
    logger.info(
        "Start waiting for deployment %s to finish.",
        latest_deployment_arn,
    )
    ok, status = wait_for_deployment_complete(latest_deployment_arn, polling_interval=polling_interval)
    if ok:
        logger.info(
            "Deployment succeeded with status %s",
            status,
        )
    else:
        msg = f"Deployment failed with status: {status}"
        raise DeploymentFailedError(msg)

    # Wait for the service to be stable
    if wait_for_stability:
        logger.info(
            "Start waiting for service %s to be stable.",
            service_ref.service,
        )
        wait_for_service_stability(service_ref, polling_interval=polling_interval)

    # Check if the service task definition matches the expected one
    if expected_task_definition:
        logger.info(
            "Checking if the service task definition is the expected one: %s",
            expected_task_definition,
        )
        ok, actual = check_service_task_definition(service_ref, expect=expected_task_definition)
        if not ok:
            msg = f"The service task definition is not the expected one; got: {actual!r}"
            raise ServiceTaskDefinitionAssertionError(msg)

        logger.info("The service task definition matches the expected one.")



from __future__ import annotations

import logging
import os

import typer

from aws_annoying.utils.ec2 import get_instance_id_by_name

from ._app import session_manager_app
from ._common import SessionManager

logger = logging.getLogger(__name__)

# TODO(lasuillard): ECS support (#24)
# TODO(lasuillard): Interactive instance selection


@session_manager_app.command()
def start(
    ctx: typer.Context,
    *,
    target: str = typer.Option(
        ...,
        show_default=False,
        help="The name or ID of the EC2 instance to connect to.",
    ),
    reason: str = typer.Option(
        "",
        help="The reason for starting the session.",
    ),
) -> None:
    """Start new session."""
    dry_run = ctx.meta["dry_run"]
    session_manager = SessionManager()

    # Resolve the instance name or ID
    instance_id = get_instance_id_by_name(target)
    if instance_id:
        logger.info("Instance ID resolved: [bold]%s[/bold]", instance_id)
        target = instance_id
    else:
        logger.info("Instance with name '%s' not found.", target)
        raise typer.Exit(1)

    # Start the session, replacing the current process
    logger.info(
        "Starting session to target [bold]%s[/bold] with reason: [italic]%r[/italic].",
        target,
        reason,
    )
    command = session_manager.build_command(
        target=target,
        document_name="SSM-SessionManagerRunShell",
        parameters={},
        reason=reason,
    )
    if not dry_run:
        os.execvp(command[0], command)  # noqa: S606



from __future__ import annotations

import logging
from pathlib import Path  # noqa: TC003
from typing import Optional

import boto3
import typer
from rich.prompt import Prompt

from aws_annoying.mfa_config import MfaConfig, update_credentials

from ._app import mfa_app

logger = logging.getLogger(__name__)


@mfa_app.command()
def configure(  # noqa: PLR0913
    ctx: typer.Context,
    *,
    mfa_profile: Optional[str] = typer.Option(
        None,
        help="The MFA profile to configure.",
    ),
    mfa_source_profile: Optional[str] = typer.Option(
        None,
        help="The AWS profile to use to retrieve MFA credentials.",
    ),
    mfa_serial_number: Optional[str] = typer.Option(
        None,
        help="The MFA device serial number. It is required if not persisted in configuration.",
        show_default=False,
    ),
    mfa_token_code: Optional[str] = typer.Option(
        None,
        help="The MFA token code.",
        show_default=False,
    ),
    aws_credentials: Path = typer.Option(  # noqa: B008
        "~/.aws/credentials",
        help="The path to the AWS credentials file.",
    ),
    aws_config: Path = typer.Option(  # noqa: B008
        "~/.aws/config",
        help="The path to the AWS config file. Used to persist the MFA configuration.",
    ),
    aws_config_section: str = typer.Option(
        "aws-annoying:mfa",
        help="The section in the AWS config file to persist the MFA configuration.",
    ),
    persist: bool = typer.Option(
        True,  # noqa: FBT003
        help="Persist the MFA configuration.",
    ),
) -> None:
    """Configure AWS profile for MFA."""
    dry_run = ctx.meta["dry_run"]

    # Expand user home directory
    aws_credentials = aws_credentials.expanduser()
    aws_config = aws_config.expanduser()

    # Load configuration
    mfa_config, exists = MfaConfig.from_ini_file(aws_config, aws_config_section)
    if exists:
        logger.info("Loaded MFA configuration from AWS config (%s).", aws_config)

    mfa_profile = (
        mfa_profile
        or mfa_config.mfa_profile
        # _
        or Prompt.ask("👤 Enter name of MFA profile to configure", default="mfa")
    )
    mfa_source_profile = (
        mfa_source_profile
        or mfa_config.mfa_source_profile
        or Prompt.ask("👤 Enter AWS profile to use to retrieve MFA credentials", default="default")
    )
    mfa_serial_number = (
        mfa_serial_number
        or mfa_config.mfa_serial_number
        # _
        or Prompt.ask("🔒 Enter MFA serial number")
    )
    mfa_token_code = (
        mfa_token_code
        # _
        or Prompt.ask("🔑 Enter MFA token code")
    )

    # Get credentials
    logger.info("Retrieving MFA credentials using profile [bold]%s[/bold]", mfa_source_profile)
    session = boto3.session.Session(profile_name=mfa_source_profile)
    sts = session.client("sts")
    response = sts.get_session_token(
        SerialNumber=mfa_serial_number,
        TokenCode=mfa_token_code,
    )
    credentials = response["Credentials"]

    # Update MFA profile in AWS credentials
    logger.warning(
        "Updating MFA profile ([bold]%s[/bold]) to AWS credentials ([bold]%s[/bold])",
        mfa_profile,
        aws_credentials,
    )
    if not dry_run:
        update_credentials(
            aws_credentials,
            mfa_profile,  # type: ignore[arg-type]
            access_key=credentials["AccessKeyId"],
            secret_key=credentials["SecretAccessKey"],
            session_token=credentials["SessionToken"],
        )

    # Persist MFA configuration
    if persist:
        logger.info(
            "Persisting MFA configuration in AWS config (%s), in [bold]%s[/bold] section.",
            aws_config,
            aws_config_section,
        )
        mfa_config.mfa_profile = mfa_profile
        mfa_config.mfa_source_profile = mfa_source_profile
        mfa_config.mfa_serial_number = mfa_serial_number
        if not dry_run:
            mfa_config.save_ini_file(aws_config, aws_config_section)
    else:
        logger.warning("MFA configuration not persisted.")



from __future__ import annotations

from datetime import datetime
from unittest import mock

import boto3
import pytest
from botocore.stub import Stubber

from aws_annoying.ecs import (
    ECSServiceRef,
    NoRunningDeploymentError,
    wait_for_deployment_complete,
    wait_for_deployment_start,
    wait_for_service_stability,
)

pytestmark = [
    pytest.mark.unit,
]


class Test_wait_for_deployment_start:
    def test_wait_for_deployment_start(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        for _ in range(2):
            stubber.add_response(
                "list_service_deployments",
                {"serviceDeployments": []},
                expected_params={
                    "cluster": "my-cluster",
                    "service": "my-service",
                    "status": ["PENDING", "IN_PROGRESS"],
                },
            )

        stubber.add_response(
            "list_service_deployments",
            {
                "serviceDeployments": [
                    {
                        "serviceDeploymentArn": "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",  # noqa: E501
                        "startedAt": datetime(2025, 5, 22, 17, 59, 58, 808000),  # noqa: DTZ001
                        "status": "PENDING",
                    },
                ],
            },
            expected_params={
                "cluster": "my-cluster",
                "service": "my-service",
                "status": ["PENDING", "IN_PROGRESS"],
            },
        )

        # Act & Assert
        with stubber:
            assert (
                wait_for_deployment_start(
                    ECSServiceRef(cluster="my-cluster", service="my-service"),
                    session=mocked_session,
                    wait_for_start=True,
                    polling_interval=1,
                    max_attempts=3,
                )
                == "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1"  # noqa: E501
            )

    def test_wait_for_deployment_start_no_deployment(self) -> None:
        """If there is no deployment, it should raise `NoRunningDeploymentError`."""
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "list_service_deployments",
            {"serviceDeployments": []},
            expected_params={
                "cluster": "my-cluster",
                "service": "my-service",
                "status": ["PENDING", "IN_PROGRESS"],
            },
        )

        # Act & Assert
        with stubber, pytest.raises(NoRunningDeploymentError):
            wait_for_deployment_start(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                session=mocked_session,
                wait_for_start=False,
            )

    def test_wait_for_deployment_start_max_attempts_exceeded(self) -> None:
        """If there is no deployment after max attempts, it should raise `NoRunningDeploymentError`."""
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        for _ in range(5):
            stubber.add_response(
                "list_service_deployments",
                {"serviceDeployments": []},
                expected_params={
                    "cluster": "my-cluster",
                    "service": "my-service",
                    "status": ["PENDING", "IN_PROGRESS"],
                },
            )

        # Act & Assert
        with stubber, pytest.raises(NoRunningDeploymentError):
            wait_for_deployment_start(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                session=mocked_session,
                wait_for_start=True,
                polling_interval=1,
                max_attempts=3,
            )


class Test_wait_for_deployment_complete:
    def test_wait_for_deployment_complete(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "describe_service_deployments",
            {"serviceDeployments": [{"status": "PENDING"}]},
            expected_params={
                "serviceDeploymentArns": [
                    "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                ],
            },
        )
        stubber.add_response(
            "describe_service_deployments",
            {"serviceDeployments": [{"status": "IN_PROGRESS"}]},
            expected_params={
                "serviceDeploymentArns": [
                    "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                ],
            },
        )
        stubber.add_response(
            "describe_service_deployments",
            {"serviceDeployments": [{"status": "SUCCESSFUL"}]},
            expected_params={
                "serviceDeploymentArns": [
                    "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                ],
            },
        )

        # Act
        with stubber:
            ok, actual = wait_for_deployment_complete(
                "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                session=mocked_session,
                polling_interval=1,
                max_attempts=3,
            )

        # Assert
        assert ok is True
        assert actual == "SUCCESSFUL"

    def test_wait_for_deployment_complete_max_attempts_exceeded(self) -> None:
        """If the deployment is still in incomplete status after max attempts, it should return `False` and last status."""  # noqa: E501
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "describe_service_deployments",
            {"serviceDeployments": [{"status": "PENDING"}]},
            expected_params={
                "serviceDeploymentArns": [
                    "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                ],
            },
        )
        for _ in range(4):
            stubber.add_response(
                "describe_service_deployments",
                {"serviceDeployments": [{"status": "IN_PROGRESS"}]},
                expected_params={
                    "serviceDeploymentArns": [
                        "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                    ],
                },
            )

        # Act
        with stubber:
            ok, actual = wait_for_deployment_complete(
                "arn:aws:ecs:ap-northeast-2:000000000000:service-deployment/my-cluster/my-service/wAMeGIKKhxAmoq1Ef03r1",
                session=mocked_session,
                polling_interval=1,
                max_attempts=3,
            )

        # Assert
        assert ok is False
        assert actual == "IN_PROGRESS"

    @pytest.mark.parametrize(
        "status",
        [
            "STOPPED",
            "STOP_REQUESTED",
            "ROLLBACK_REQUESTED",
            "ROLLBACK_IN_PROGRESS",
            "ROLLBACK_SUCCESSFUL",
            "ROLLBACK_FAILED",
        ],
    )
    def test_wait_for_deployment_complete_failed(self, status: str) -> None:
        """If the deployment is in a failed status, it should return `False` with the status."""
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        for _ in range(2):
            stubber.add_response(
                "describe_service_deployments",
                {"serviceDeployments": [{"status": "IN_PROGRESS"}]},
                expected_params={
                    "serviceDeploymentArns": [
                        "arn:aws:ecs:us-east-1:123456789012:service-deployment/example-cluster/example-service/ejGvqq2ilnbKT9qj0vLJe",
                    ],
                },
            )

        stubber.add_response(
            "describe_service_deployments",
            {"serviceDeployments": [{"status": status}]},
            expected_params={
                "serviceDeploymentArns": [
                    "arn:aws:ecs:us-east-1:123456789012:service-deployment/example-cluster/example-service/ejGvqq2ilnbKT9qj0vLJe",
                ],
            },
        )

        # Act
        with stubber:
            ok, actual = wait_for_deployment_complete(
                "arn:aws:ecs:us-east-1:123456789012:service-deployment/example-cluster/example-service/ejGvqq2ilnbKT9qj0vLJe",
                session=mocked_session,
                polling_interval=1,
                max_attempts=3,
            )

        # Assert
        assert ok is False
        assert actual == status


class Test_wait_for_service_stability:
    def test_wait_for_service_stability(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        stubber.add_response(
            "describe_services",
            {
                "services": [
                    {
                        "status": "ACTIVE",
                        "desiredCount": 1,
                        "runningCount": 0,
                        "deployments": [
                            # ...
                            {},
                        ],
                    },
                ],
            },
            expected_params={"cluster": "my-cluster", "services": ["my-service"]},
        )
        for _ in range(2):
            stubber.add_response(
                "describe_services",
                {
                    "services": [
                        {
                            "desiredCount": 1,
                            "runningCount": 1,
                            "deployments": [
                                # ...
                                {},
                            ],
                        },
                    ],
                },
                expected_params={"cluster": "my-cluster", "services": ["my-service"]},
            )

        # Act
        with stubber:
            ok = wait_for_service_stability(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                session=mocked_session,
                polling_interval=1,
                max_attempts=3,
            )

        # Assert
        assert ok is True

    def test_wait_for_service_stability_max_attempts_exceeded(self) -> None:
        # Arrange
        mocked_session = mock.MagicMock()
        mocked_session.client.return_value = ecs = boto3.client("ecs")
        stubber = Stubber(ecs)
        for _ in range(5):
            stubber.add_response(
                "describe_services",
                {
                    "services": [
                        {
                            "desiredCount": 1,
                            "runningCount": 0,
                            "deployments": [
                                # ...
                                {},
                            ],
                        },
                    ],
                },
                expected_params={"cluster": "my-cluster", "services": ["my-service"]},
            )

        # Act
        with stubber:
            ok = wait_for_service_stability(
                ECSServiceRef(cluster="my-cluster", service="my-service"),
                session=mocked_session,
                polling_interval=1,
                max_attempts=3,
            )

        # Assert
        assert ok is False



from itertools import chain


def repeat_options(option: str, values: list[str]) -> list[str]:
    """Repeat an option for each value."""
    return list(chain.from_iterable((option, value) for value in values))



from .check import check_service_task_definition
from .common import ECSServiceRef
from .errors import (
    DeploymentFailedError,
    NoRunningDeploymentError,
    ServiceTaskDefinitionAssertionError,
    WaitForDeploymentError,
)
from .wait_for import (
    wait_for_deployment_complete,
    wait_for_deployment_start,
    wait_for_service_stability,
)

__all__ = (
    "DeploymentFailedError",
    "ECSServiceRef",
    "NoRunningDeploymentError",
    "ServiceTaskDefinitionAssertionError",
    "WaitForDeploymentError",
    "check_service_task_definition",
    "wait_for_deployment_complete",
    "wait_for_deployment_start",
    "wait_for_service_stability",
)



from __future__ import annotations

import logging
from datetime import datetime, timezone
from time import sleep
from typing import TYPE_CHECKING

import boto3
import botocore.exceptions

from .errors import NoRunningDeploymentError

if TYPE_CHECKING:
    from .common import ECSServiceRef

logger = logging.getLogger(__name__)


def wait_for_deployment_start(
    service_ref: ECSServiceRef,
    *,
    session: boto3.session.Session | None = None,
    wait_for_start: bool,
    polling_interval: int = 5,
    max_attempts: int | None = None,
) -> str:
    """Wait for the ECS deployment to start.

    Args:
        service_ref: The ECS service reference containing the cluster and service names.
        session: The boto3 session to use for the ECS client.
        wait_for_start: Whether to wait for the deployment to start.
        polling_interval: The interval between any polling attempts, in seconds.
        max_attempts: The maximum number of attempts to wait for the deployment to start.

    Raises:
        NoRunningDeploymentError: If no running deployments are found and `wait_for_start` is False.

    Returns:
        The ARN of the latest deployment for the service.
    """
    session = session or boto3.session.Session()
    ecs = session.client("ecs")

    if wait_for_start:
        logger.warning("`wait_for_start` is set, will wait for a new deployment to start.")

    attempts = 0
    while True:  # do-while
        # Do
        running_deployments = ecs.list_service_deployments(
            cluster=service_ref.cluster,
            service=service_ref.service,
            status=["PENDING", "IN_PROGRESS"],
        )["serviceDeployments"]

        # While
        if running_deployments:
            logger.debug("Found %d running deployments for service. Exiting loop.", len(running_deployments))
            break

        if not wait_for_start:
            logger.debug("`wait_for_start` is off, no need to wait for a new deployment to start.")
            break

        if max_attempts and attempts >= max_attempts:
            logger.debug("Max attempts exceeded while waiting for a new deployment to start.")
            break

        logger.debug(
            "(%d-th attempt) No running deployments found for service. Start waiting for a new deployment.",
            attempts + 1,
        )

        sleep(polling_interval)
        attempts += 1

    if not running_deployments:
        msg = "No running deployments found for service."
        raise NoRunningDeploymentError(msg)

    latest_deployment = max(
        running_deployments,
        key=lambda dep: dep.get(
            "startedAt",
            datetime.min.replace(tzinfo=timezone.utc),
        ),
    )
    if len(running_deployments) > 1:
        logger.warning(
            "%d running deployments found for service. Using most recently started deployment: %s",
            len(running_deployments),
            latest_deployment["serviceDeploymentArn"],
        )

    return latest_deployment["serviceDeploymentArn"]


def wait_for_deployment_complete(
    deployment_arn: str,
    *,
    session: boto3.session.Session | None = None,
    polling_interval: int = 5,
    max_attempts: int | None = None,
) -> tuple[bool, str]:
    """Wait for the ECS deployment to complete.

    Args:
        deployment_arn: The ARN of the deployment to wait for.
        session: The boto3 session to use for the ECS client.
        polling_interval: The interval between any polling attempts, in seconds.
        max_attempts: The maximum number of attempts to wait for the deployment to complete.

    Returns:
        A tuple containing a boolean indicating whether the deployment succeeded and the status of the deployment.
    """
    session = session or boto3.session.Session()
    ecs = session.client("ecs")

    attempts = 0
    while (max_attempts is None) or (attempts <= max_attempts):
        latest_deployment = ecs.describe_service_deployments(serviceDeploymentArns=[deployment_arn])[
            "serviceDeployments"
        ][0]
        status = latest_deployment["status"]
        if status == "SUCCESSFUL":
            return (True, status)

        if status in ("PENDING", "IN_PROGRESS"):
            logger.debug(
                "(%d-th attempt) Deployment in progress... (%s)",
                attempts + 1,
                status,
            )
        else:
            break

        sleep(polling_interval)
        attempts += 1

    return (False, status)


def wait_for_service_stability(
    service_ref: ECSServiceRef,
    *,
    session: boto3.session.Session | None = None,
    polling_interval: int = 5,
    max_attempts: int | None = None,
) -> bool:
    """Wait for the ECS service to be stable.

    Args:
        service_ref: The ECS service reference containing the cluster and service names.
        session: The boto3 session to use for the ECS client.
        polling_interval: The interval between any polling attempts, in seconds.
        max_attempts: The maximum number of attempts to wait for the service to be stable.

    Returns:
        A boolean indicating whether the service is stable.
    """
    session = session or boto3.session.Session()
    ecs = session.client("ecs")

    # TODO(lasuillard): Likely to be a problem in some cases: https://github.com/boto/botocore/issues/3314
    stability_waiter = ecs.get_waiter("services_stable")

    attempts = 0
    while (max_attempts is None) or (attempts <= max_attempts):
        logger.debug(
            "(%d-th attempt) Waiting for service %s to be stable...",
            attempts + 1,
            service_ref.service,
        )
        try:
            stability_waiter.wait(
                cluster=service_ref.cluster,
                services=[service_ref.service],
                WaiterConfig={"Delay": polling_interval, "MaxAttempts": 1},
            )
        except botocore.exceptions.WaiterError as err:
            if err.kwargs["reason"] != "Max attempts exceeded":
                raise
        else:
            return True

        sleep(polling_interval)
        attempts += 1

    return False



class SessionManagerError(Exception):
    """Base exception for all errors related to Session Manager."""


class UnsupportedPlatformError(SessionManagerError):
    """Exception raised when the platform is not supported."""


class PluginNotInstalledError(SessionManagerError):
    """Trying to use the Session Manager plugin before it is installed."""



import typer

from aws_annoying.cli.app import app

session_manager_app = typer.Typer(
    no_args_is_help=True,
    help="AWS Session Manager CLI utilities.",
)
app.add_typer(session_manager_app, name="session-manager")



from __future__ import annotations

from time import sleep

import pytest

from aws_annoying.utils.timeout import OperationTimeoutError, Timeout
from tests._helpers import run_if_windows, skip_if_windows


class Test_Timeout:
    @skip_if_windows
    def test_decorator_non_windows(self) -> None:
        # Arrange
        @Timeout(1)
        def do_something() -> None:
            sleep(3)

        # Act & Assert
        with pytest.raises(OperationTimeoutError):
            do_something()

    @skip_if_windows
    def test_context_manager_non_windows(self) -> None:
        # Arrange & Act & Assert
        with Timeout(1), pytest.raises(OperationTimeoutError):
            sleep(3)

    @run_if_windows
    def test_decorator_windows(self) -> None:
        # Arrange
        @Timeout(1)
        def do_something() -> None:
            sleep(3)

        # Act & Assert
        do_something()  # Won't raise an exception

    @run_if_windows
    def test_context_manager_windows(self) -> None:
        # Arrange & Act & Assert
        with Timeout(1):
            sleep(3)  # Won't raise an exception



#!/usr/bin/env python
"""Test helper script to print environment variables.

This script is slightly different from `printenv` command in Unix-like systems.

- It prints the environment variables' with the provided keys only.
- If variable is not found, it prints an empty string to easily compare with the expected value.
"""

import argparse
import json
import os
import sys


def main() -> None:
    """Print the environment variables."""
    parser = argparse.ArgumentParser(description="Print the environment variables.")
    parser.add_argument("--json", action="store_true", help="Print the environment variables as JSON.", default=False)
    parser.add_argument("keys", nargs="*", help="The keys to print.")
    ns = parser.parse_args()

    # Arguments
    keys: list[str] = ns.keys
    as_json: bool = ns.json

    # Print the environment variables
    env = {key: os.environ.get(key, "") for key in keys}
    if as_json:
        sys.stdout.write(f"{json.dumps(env)}\n")
    else:
        sys.stdout.writelines(f"{k}={v}\n" for k, v in env.items())


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import os
import signal
import subprocess
from pathlib import Path  # noqa: TC003

import typer

from aws_annoying.utils.ec2 import get_instance_id_by_name

from ._app import session_manager_app
from ._common import SessionManager

logger = logging.getLogger(__name__)


# https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html
@session_manager_app.command()
def port_forward(  # noqa: PLR0913
    ctx: typer.Context,
    *,
    # TODO(lasuillard): Add `--local-host` option, redirect the traffic to non-localhost bind (unsupported by AWS)
    local_port: int = typer.Option(
        ...,
        show_default=False,
        help="The local port to use for port forwarding.",
    ),
    through: str = typer.Option(
        ...,
        show_default=False,
        help="The name or ID of the EC2 instance to use as a proxy for port forwarding.",
    ),
    remote_host: str = typer.Option(
        ...,
        show_default=False,
        help="The remote host to connect to.",
    ),
    remote_port: int = typer.Option(
        ...,
        show_default=False,
        help="The remote port to connect to.",
    ),
    reason: str = typer.Option(
        "",
        help="The reason for starting the port forwarding session.",
    ),
    pid_file: Path = typer.Option(  # noqa: B008
        "./session-manager-plugin.pid",
        help="The path to the PID file to store the process ID of the session manager plugin.",
    ),
    terminate_running_process: bool = typer.Option(
        False,  # noqa: FBT003
        help="Terminate the process in the PID file if it already exists.",
    ),
    log_file: Path = typer.Option(  # noqa: B008
        "./session-manager-plugin.log",
        help="The path to the log file to store the output of the session manager plugin.",
    ),
) -> None:
    """Start a port forwarding session using AWS Session Manager."""
    dry_run = ctx.meta["dry_run"]
    session_manager = SessionManager()

    # Check if the PID file already exists
    if pid_file.exists():
        if not terminate_running_process:
            logger.error("PID file already exists.")
            raise typer.Exit(1)

        pid_content = pid_file.read_text()
        try:
            existing_pid = int(pid_content)
        except ValueError:
            logger.error("PID file content is invalid; expected integer, but got: %r", type(pid_content))  # noqa: TRY400
            raise typer.Exit(1) from None

        try:
            logger.warning("Terminating running process with PID %d.", existing_pid)
            os.kill(existing_pid, signal.SIGTERM)
            pid_file.write_text("")  # Clear the PID file
        except ProcessLookupError:
            logger.warning("Tried to terminate process with PID %d but does not exist.", existing_pid)

    # Resolve the instance name or ID
    instance_id = get_instance_id_by_name(through)
    if instance_id:
        logger.info("Instance ID resolved: [bold]%s[/bold]", instance_id)
        target = instance_id
    else:
        logger.error("Instance with name '%s' not found.", through)
        raise typer.Exit(1)

    # Initiate the session
    command = session_manager.build_command(
        target=target,
        document_name="AWS-StartPortForwardingSessionToRemoteHost",
        parameters={
            "host": [remote_host],
            "portNumber": [str(remote_port)],
            "localPortNumber": [str(local_port)],
        },
        reason=reason,
    )
    stdout: subprocess._FILE
    if log_file is not None:  # noqa: SIM108
        stdout = log_file.open(mode="at+", buffering=1)
    else:
        stdout = subprocess.DEVNULL

    logger.info(
        "Starting port forwarding session through [bold]%s[/bold] with reason: [italic]%r[/italic].",
        through,
        reason,
    )
    if not dry_run:
        proc = subprocess.Popen(  # noqa: S603
            command,
            stdout=stdout,
            stderr=subprocess.STDOUT,
            text=True,
            close_fds=False,  # FD inherited from parent process
        )
        pid = proc.pid
    else:
        pid = -1

    logger.info(
        "Session Manager Plugin started with PID %d. Outputs will be logged to %s.",
        pid,
        log_file.absolute(),
    )

    # Write the PID to the file
    if not dry_run:
        pid_file.write_text(str(pid))

    logger.info("PID file written to %s.", pid_file.absolute())



from __future__ import annotations

import logging

import typer

from aws_annoying.utils.downloader import TQDMDownloader

from ._app import session_manager_app
from ._common import SessionManager

logger = logging.getLogger(__name__)


# https://docs.aws.amazon.com/systems-manager/latest/userguide/session-manager-working-with-install-plugin.html
@session_manager_app.command()
def install(
    ctx: typer.Context,
    *,
    yes: bool = typer.Option(
        False,  # noqa: FBT003
        help="Do not ask confirmation for installation.",
    ),
) -> None:
    """Install AWS Session Manager plugin."""
    dry_run = ctx.meta["dry_run"]
    session_manager = SessionManager()

    # Check session-manager-plugin already installed
    is_installed, binary_path, version = session_manager.verify_installation()
    if is_installed:
        logger.info("Session Manager plugin is already installed at %s (version: %s)", binary_path, version)
        return

    # Install session-manager-plugin
    logger.warning("Installing AWS Session Manager plugin. You could be prompted for admin privileges request.")
    if not dry_run:
        session_manager.install(confirm=yes, downloader=TQDMDownloader())

    # Verify installation
    is_installed, binary_path, version = session_manager.verify_installation()
    if not is_installed:
        logger.error("Installation failed. Session Manager plugin not found.")
        raise typer.Exit(1)

    logger.info("Session Manager plugin successfully installed at %s (version: %s)", binary_path, version)



class WaitForDeploymentError(Exception):
    """Base class for all deployment waiter errors."""


class NoRunningDeploymentError(WaitForDeploymentError):
    """No running deployment found for the service."""


class DeploymentFailedError(WaitForDeploymentError):
    """Deployment failed."""


class ServiceTaskDefinitionAssertionError(WaitForDeploymentError):
    """Service task definition does not match the expected one."""



# flake8: noqa: B008
from __future__ import annotations

import json
import logging
from typing import Any, TypedDict

import boto3

logger = logging.getLogger(__name__)

# Type aliases for readability
_ARN = str
_Variables = dict[str, Any]


class _LoadStatsDict(TypedDict):
    secrets: int
    parameters: int


class VariableLoader:  # noqa: D101
    def __init__(self, *, session: boto3.session.Session | None = None) -> None:
        """Initialize variable loader.

        Args:
            session: Boto3 session to use for AWS operations.
        """
        self.session = session or boto3.session.Session()

    # TODO(lasuillard): Currently not using pagination (do we need more than 10-20 secrets or parameters each?)
    #                   ; consider adding it if needed
    def load(self, map_arns: dict[str, _ARN]) -> tuple[dict[str, Any], _LoadStatsDict]:
        """Load the variables from the AWS Secrets Manager and SSM Parameter Store.

        Each secret or parameter should be a valid dictionary, where the keys are the variable names
        and the values are the variable values.

        The items are merged in the order of the key of provided mapping, overwriting the variables with the same name
        in the order of the keys.
        """
        # Split the ARNs by resource types
        secrets_map, parameters_map = {}, {}
        for idx, arn in map_arns.items():
            if arn.startswith("arn:aws:secretsmanager:"):
                secrets_map[idx] = arn
            elif arn.startswith("arn:aws:ssm:"):
                parameters_map[idx] = arn
            else:
                msg = f"Unsupported resource: {arn!r}"
                raise ValueError(msg)

        # Retrieve variables from AWS resources
        secrets: dict[str, _Variables]
        parameters: dict[str, _Variables]
        secrets = self._retrieve_secrets(secrets_map)
        parameters = self._retrieve_parameters(parameters_map)

        load_stats: _LoadStatsDict = {
            "secrets": len(secrets),
            "parameters": len(parameters),
        }

        # Merge the variables in order
        full_variables = secrets | parameters  # Keys MUST NOT conflict
        merged_in_order = {}
        for _, variables in sorted(full_variables.items()):
            merged_in_order.update(variables)

        return merged_in_order, load_stats

    def _retrieve_secrets(self, secrets_map: dict[str, _ARN]) -> dict[str, _Variables]:
        """Retrieve the secrets from AWS Secrets Manager."""
        if not secrets_map:
            return {}

        secretsmanager = self.session.client("secretsmanager")

        # Retrieve the secrets
        arns = list(secrets_map.values())
        response = secretsmanager.batch_get_secret_value(SecretIdList=arns)
        if errors := response["Errors"]:
            msg = f"Failed to retrieve secrets: {errors!r}"
            raise ValueError(msg)

        # Parse the secrets
        secrets = response["SecretValues"]
        result = {}
        for secret in secrets:
            arn = secret["ARN"]
            order_key = next(key for key, value in secrets_map.items() if value == arn)
            data = json.loads(secret["SecretString"])
            if not isinstance(data, dict):
                msg = f"Secret data must be a valid dictionary, but got: {type(data)!r}"
                raise TypeError(msg)

            result[order_key] = data

        return result

    def _retrieve_parameters(self, parameters_map: dict[str, _ARN]) -> dict[str, _Variables]:
        """Retrieve the parameters from AWS SSM Parameter Store."""
        if not parameters_map:
            return {}

        ssm = self.session.client("ssm")

        # Retrieve the parameters
        parameter_names = list(parameters_map.values())
        response = ssm.get_parameters(Names=parameter_names, WithDecryption=True)
        if errors := response["InvalidParameters"]:
            msg = f"Failed to retrieve parameters: {errors!r}"
            raise ValueError(msg)

        # Parse the parameters
        parameters = response["Parameters"]
        result = {}
        for parameter in parameters:
            arn = parameter["ARN"]
            order_key = next(key for key, value in parameters_map.items() if value == arn)
            data = json.loads(parameter["Value"])
            if not isinstance(data, dict):
                msg = f"Parameter data must be a valid dictionary, but got: {type(data)!r}"
                raise TypeError(msg)

            result[order_key] = data

        return result



from __future__ import annotations

import logging
import logging.config
from typing import TYPE_CHECKING, Any, Final

from typing_extensions import override

if TYPE_CHECKING:
    from rich.console import Console


class RichLogHandler(logging.Handler):
    """Custom logging handler to use Rich Console."""

    _level_emojis: Final[dict[str, str]] = {
        "DEBUG": "🔍",
        "INFO": "🔔",
        "WARNING": "⚠️",
        "ERROR": "🚨",
        "CRITICAL": "🔥",
    }

    def __init__(self, console: Console, *args: Any, **kwargs: Any) -> None:
        """Initialize the log handler.

        Args:
            console: Rich console instance.
            *args: Additional arguments for the logging handler.
            **kwargs: Additional keyword arguments for the logging handler.
        """
        super().__init__(*args, **kwargs)
        self.console = console

    @override
    def emit(self, record: logging.LogRecord) -> None:
        msg = self.format(record)
        self.console.print(msg)

    @override
    def format(self, record: logging.LogRecord) -> str:
        """Format the log record.

        Args:
            record: The log record to format.

        Returns:
            The formatted log message.
        """
        msg = super().format(record)
        emoji = self._level_emojis.get(record.levelname)
        return f"{emoji} {msg}" if emoji else msg



from __future__ import annotations

import re

import boto3


def get_instance_id_by_name(name_or_id: str, *, session: boto3.session.Session | None = None) -> str | None:
    """Get the EC2 instance ID by name or ID.

    Be aware that this function will only return the first instance found
    with the given name, no matter how many instances are found.

    Args:
        name_or_id: The name or ID of the EC2 instance.
        session: The boto3 session to use. If not provided, a new session will be created.

    Returns:
        The instance ID if found, otherwise `None`.
    """
    if re.match(r"^m?i-[0-9a-f]+$", name_or_id):
        return name_or_id

    session = session or boto3.session.Session()
    ec2 = session.client("ec2")

    response = ec2.describe_instances(Filters=[{"Name": "tag:Name", "Values": [name_or_id]}])
    reservations = response["Reservations"]
    if not reservations or not reservations[0]["Instances"]:
        return None

    instances = reservations[0]["Instances"]
    return str(instances[0]["InstanceId"])



from __future__ import annotations

from pathlib import Path

import pytest
import toml
from typer.testing import CliRunner

from aws_annoying.cli.main import app

runner = CliRunner()

pytestmark = [
    pytest.mark.unit,
]


def test_version() -> None:
    """Show version with `--version` top-level argument."""
    # Arrange
    pyproject = toml.load(Path(__file__).parent.parent.parent / "pyproject.toml")
    current_version = pyproject["project"]["version"]

    # Act
    result = runner.invoke(
        app,
        [
            "--version",
        ],
    )

    # Assert
    assert result.exit_code == 0
    assert result.stdout.strip() == current_version



from .errors import PluginNotInstalledError, SessionManagerError, UnsupportedPlatformError
from .session_manager import SessionManager
from .shortcuts import port_forward

__all__ = (
    "PluginNotInstalledError",
    "SessionManager",
    "SessionManagerError",
    "UnsupportedPlatformError",
    "port_forward",
)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import boto3
import typer

from ._app import ecs_app

if TYPE_CHECKING:
    from collections.abc import Iterator


logger = logging.getLogger(__name__)

_DELETE_CHUNK_SIZE = 10


@ecs_app.command()
def task_definition_lifecycle(
    ctx: typer.Context,
    *,
    family: str = typer.Option(
        ...,
        help="The name of the task definition family.",
        show_default=False,
    ),
    keep_latest: int = typer.Option(
        ...,
        help="Number of latest (revision) task definitions to keep.",
        show_default=False,
        min=1,
        max=100,
    ),
    delete: bool = typer.Option(
        False,  # noqa: FBT003
        help="Delete the task definition after deregistering it.",
    ),
) -> None:
    """Execute ECS task definition lifecycle."""
    dry_run = ctx.meta["dry_run"]
    ecs = boto3.client("ecs")

    # Get all task definitions for the family
    response_iter = ecs.get_paginator("list_task_definitions").paginate(
        familyPrefix=family,
        status="ACTIVE",
        sort="ASC",
    )
    task_definition_arns = []
    for response in response_iter:
        task_definition_arns.extend(response["taskDefinitionArns"])

    # Sort by revision number
    task_definition_arns.sort(key=lambda arn: int(arn.split(":")[-1]))

    # Keep the latest N task definitions
    expired_taskdef_arns = task_definition_arns[:-keep_latest]
    logger.warning("Deregistering %d task definitions...", len(expired_taskdef_arns))
    for arn in expired_taskdef_arns:
        if not dry_run:
            ecs.deregister_task_definition(taskDefinition=arn)

        # ARN like: "arn:aws:ecs:<region>:<account-id>:task-definition/<family>:<revision>"
        _, family_revision = arn.split(":task-definition/")
        logger.warning("Deregistered task definition [yellow]%r[/yellow]", family_revision)

    if delete and expired_taskdef_arns:
        # Delete the expired task definitions in chunks due to API limitation
        logger.warning(
            "Deleting %d task definitions in chunks of size %d...",
            len(expired_taskdef_arns),
            _DELETE_CHUNK_SIZE,
        )
        for idx, chunk in enumerate(_chunker(expired_taskdef_arns, _DELETE_CHUNK_SIZE)):
            if not dry_run:
                ecs.delete_task_definitions(taskDefinitions=chunk)

            logger.warning("Deleted %d task definitions in %d-th batch.", len(chunk), idx)


def _chunker(sequence: list, size: int) -> Iterator[list]:
    """Yield successive chunks of a given size from the sequence."""
    for i in range(0, len(sequence), size):
        yield sequence[i : i + size]
